#include "FTPClientApp.h"


#define MSGKIND_CONNECT  0
#define MSGKIND_SEND     1
#define MSGKIND_RECORD   2
#define MSGKIND_STOP     3


Define_Module(FTPClientApp);

FTPClientApp::FTPClientApp()
{
    timeoutMsg = NULL;
}

FTPClientApp::~FTPClientApp()
{
    cancelAndDelete(timeoutMsg);
}

void FTPClientApp::initialize()
{
    TCPGenericCliAppBase::initialize();

    timeoutMsg = new cMessage("timer");

	serverClose = false;

	tabIndex = 0;
	intervalSize = totalSize = 0;
	for (int k=0; k<TAB_SIZE; k++)
	{
		_intervalSizes[k] = 0;
	}

    timeoutMsg->setKind(MSGKIND_CONNECT);
    scheduleAt((simtime_t)par("startTime"), timeoutMsg);

	cMessage *recordMessage = new cMessage("Record Interval");
	recordMessage->setKind(MSGKIND_RECORD);
    scheduleAt(simTime() + WIN_DELTA, recordMessage);

	cMessage *stopMessage = new cMessage("Stop Time");
	stopMessage->setKind(MSGKIND_STOP);
    scheduleAt((simtime_t)par("stopTime") , stopMessage);

	lastRecvdMsg = 0;

	receivedRate.setName("TCP-Received rate with Filter");
}

void FTPClientApp::sendRequest()
{
     long requestLength = par("requestLength");
     long replyLength = par("replyLength");
     if (requestLength<1) requestLength=1;
     if (replyLength<1) replyLength=1;

     sendPacket(requestLength, replyLength, serverClose);
}

void FTPClientApp::handleTimer(cMessage *msg)
{
    switch (msg->kind())
    {
        case MSGKIND_CONNECT:
            EV << "starting session\n";
            connect(); // active OPEN

            // significance of earlySend: if true, data will be sent already
            // in the ACK of SYN, otherwise only in a separate packet (but still
            // immediately)
            sendRequest();
            break;

        case MSGKIND_SEND:
           sendRequest();
           // no scheduleAt(): next request will be sent when reply to this one
           // arrives (see socketDataArrived())
//		   scheduleAt(simTime(), timeoutMsg);
           break;

	   case MSGKIND_RECORD:
           recordReceivedRate(msg);
           break;

	   case MSGKIND_STOP:
           serverClose = true;
           break;
    }
}

void FTPClientApp::socketEstablished(int connId, void *ptr)
{
    TCPGenericCliAppBase::socketEstablished(connId, ptr);
    sendRequest();
	timeoutMsg->setKind(MSGKIND_SEND);
	scheduleAt(simTime()+0, timeoutMsg);

}

void FTPClientApp::socketDataArrived(int connId, void *ptr, cMessage *msg, bool urgent)
{
	intervalSize +=msg->length();
	if (lastRecvdMsg >0)
		receivedRate.record(msg->length()/(msg->arrivalTime()-lastRecvdMsg)/1000);
	lastRecvdMsg = msg->arrivalTime();

    TCPGenericCliAppBase::socketDataArrived(connId, ptr, msg, urgent);

    EV << "reply arrived\n";
	if (!serverClose)
		sendRequest();
}

void FTPClientApp::socketClosed(int connId, void *ptr)
{
    TCPGenericCliAppBase::socketClosed(connId, ptr);
}

void FTPClientApp::socketFailure(int connId, void *ptr, int code)
{
    TCPGenericCliAppBase::socketFailure(connId, ptr, code);
}

void FTPClientApp::recordReceivedRate(cMessage *msg)
{
	int index = tabIndex % (int)TAB_SIZE;
	double rvcdThroughput = 0.0;

	if (tabIndex >= TAB_SIZE)
	{
		double rvcdThroughput = 0.0;

		totalSize = totalSize + intervalSize - _intervalSizes[index];
		rvcdThroughput = totalSize / WIN_TIME_SIZE;
		receivedRate.record(rvcdThroughput/1000);
	}

	_intervalSizes[index] = intervalSize;
	intervalSize = 0;
	tabIndex++;
	 
    scheduleAt(simTime() + WIN_DELTA, msg);
}
