//
// Copyright 2004 Andras Varga
//
// This library is free software, you can redistribute it and/or modify
// it under  the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation;
// either version 2 of the License, or any later version.
// The library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
// See the GNU Lesser General Public License for more details.
//


#include "TCPBasicClientApp.h"


#define MSGKIND_CONNECT  0
#define MSGKIND_SEND     1
#define MSGKIND_RECORD   2
#define MSGKIND_STOP     3


Define_Module(TCPBasicClientApp);

TCPBasicClientApp::TCPBasicClientApp()
{
    timeoutMsg = NULL;
}

TCPBasicClientApp::~TCPBasicClientApp()
{
    cancelAndDelete(timeoutMsg);
}

void TCPBasicClientApp::initialize()
{
    TCPGenericCliAppBase::initialize();

    timeoutMsg = new cMessage("timer");

    numRequestsToSend = 0;
	timeStampLastMessageReceived = 0;
    earlySend = false;  // TBD make it parameter
	serverClose = false;

	tabIndex = 0;
	intervalSize = totalSize = 0;
	for (int k=0; k<TAB_SIZE; k++)
	{
		_intervalSizes[k] = 0;
	}

    WATCH(numRequestsToSend);
    WATCH(earlySend);

    timeoutMsg->setKind(MSGKIND_CONNECT);
    scheduleAt((simtime_t)par("startTime"), timeoutMsg);

	cMessage *recordMessage = new cMessage("Record Interval");
	recordMessage->setKind(MSGKIND_RECORD);
    scheduleAt(simTime() + WIN_DELTA, recordMessage);

	//cMessage *stopMessage = new cMessage("Stop Time");
	//stopMessage->setKind(MSGKIND_STOP);
    //scheduleAt((simtime_t)par("stopTime") , stopMessage);

	receivedRate.setName("TCP-Received rate with Filter");
}

void TCPBasicClientApp::sendRequest()
{
     EV << "sending request, " << numRequestsToSend-1 << " more to go\n";

     long requestLength = par("requestLength");
     long replyLength = par("replyLength");
     if (requestLength<1) requestLength=1;
     if (replyLength<1) replyLength=1;

	 if (!serverClose)
     sendPacket(requestLength, replyLength, serverClose);
}

void TCPBasicClientApp::handleTimer(cMessage *msg)
{
    switch (msg->kind())
    {
        case MSGKIND_CONNECT:
            EV << "starting session\n";
            connect(); // active OPEN

            // significance of earlySend: if true, data will be sent already
            // in the ACK of SYN, otherwise only in a separate packet (but still
            // immediately)
            if (earlySend)
                sendRequest();
            break;

        case MSGKIND_SEND:
           sendRequest();
           numRequestsToSend--;
           // no scheduleAt(): next request will be sent when reply to this one
           // arrives (see socketDataArrived())
           break;

	   case MSGKIND_RECORD:
           recordReceivedRate(msg);
           break;

	   case MSGKIND_STOP:
           serverClose = true;
           break;
    }
}

void TCPBasicClientApp::socketEstablished(int connId, void *ptr)
{
    TCPGenericCliAppBase::socketEstablished(connId, ptr);

    // determine number of requests in this session
    numRequestsToSend = (long) par("numRequestsPerSession");
    if (numRequestsToSend<1) numRequestsToSend=1;

    // perform first request if not already done (next one will be sent when reply arrives)
    if (!earlySend)
        sendRequest();
    numRequestsToSend--;
}

void TCPBasicClientApp::socketDataArrived(int connId, void *ptr, cMessage *msg, bool urgent)
{
	intervalSize +=msg->length();
    TCPGenericCliAppBase::socketDataArrived(connId, ptr, msg, urgent);

    if (numRequestsToSend>0)
    {
        EV << "reply arrived\n";
        timeoutMsg->setKind(MSGKIND_SEND);
        scheduleAt(simTime()+(simtime_t)par("thinkTime"), timeoutMsg);
    }
    else
    {
        EV << "reply to last request arrived, closing session\n";
        close();
    }
}

void TCPBasicClientApp::socketClosed(int connId, void *ptr)
{
    TCPGenericCliAppBase::socketClosed(connId, ptr);

    // start another session after a delay
    timeoutMsg->setKind(MSGKIND_CONNECT);
    scheduleAt(simTime()+(simtime_t)par("idleInterval"), timeoutMsg);
}

void TCPBasicClientApp::socketFailure(int connId, void *ptr, int code)
{
    TCPGenericCliAppBase::socketFailure(connId, ptr, code);

    // reconnect after a delay
    timeoutMsg->setKind(MSGKIND_CONNECT);
    scheduleAt(simTime()+(simtime_t)par("reconnectInterval"), timeoutMsg);
}

void TCPBasicClientApp::recordReceivedRate(cMessage *msg)
{
	int index = tabIndex % (int)TAB_SIZE;
	double rvcdThroughput = 0.0;

	if (tabIndex >= TAB_SIZE)
	{
		double rvcdThroughput = 0.0;

		totalSize = totalSize + intervalSize - _intervalSizes[index];
		rvcdThroughput = totalSize / WIN_TIME_SIZE;
		receivedRate.record(rvcdThroughput/1000);
	}

	_intervalSizes[index] = intervalSize;
	intervalSize = 0;
	tabIndex++;
	 
    scheduleAt(simTime() + WIN_DELTA, msg);
}
