//
// Copyright (C) 2006 Levente M�z�os
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

#ifndef IEEE80211_CONSTS_H
#define IEEE80211_CONSTS_H

// XXX these are taken over from Consts80211.h
// the phy has recognized a bit error in the packet
#define BITERROR -1
// packet lost due to collision
#define COLLISION 9

/** @brief Bit rates for 802.11a */
const double BITRATES_80211a[] = {
    6000000,
    9000000,
    12000000,
    18000000,
    24000000,
    36000000,
    48000000,
    54000000
};

/** @brief Bit rates for 802.11b */
const double BITRATES_80211b[] = {
    1000000,
    2000000,
    5500000,
    11000000
};

/** @brief Bit rates for 802.11g */
const double BITRATES_80211g[] = {
    6000000,
    9000000,
    12000000,
    18000000,
    24000000,
    36000000,
    48000000,
    54000000
};

const double RED_PHY_HEADER_DURATION = 0.000020;
/** @brief Preambule transmision delay */
const double PLCP_PREAMBLE_DELAY = 20E-6;
/** @brief Length of PLCP deader without SERVICE and Tail */
const int PLCP_HEADER_LENGTH = 4 + 1 + 12 + 1 + 6;
/** @brief SIGNAL delay */
const double PLCP_SIGNAL_DELAY = 4E-6;
/** @brief SYMBOL delay */
const double T_SYM = 4E-6;

// frame lengths in bits
// XXX this is duplicate, it's already in Ieee80211Frame.msg
const unsigned int LENGTH_RTS = 160;
const unsigned int LENGTH_CTS = 112;
const unsigned int LENGTH_ACK = 112;

const int RETRY_LIMIT = 7;

/** Minimum size (initial size) of contention window */
const int CW_MIN = 15;

/** Maximum size of contention window */
const int CW_MAX = 1023;

const int PHY_HEADER_LENGTH_A = 4 + 1 + 12 + 1 + 6 + 16 + 6;
const int PHY_HEADER_LENGTH_B = 192;
const double PHY_HEADER_LENGTH_G = 26E-6;
const int HEADER_WITHOUT_PREAMBLE = 48;
const double BITRATE_HEADER_A = 6E+6;
const double BITRATE_HEADER_B = 1E+6;
const double BANDWIDTH = 2E+6;

/** @brief BasicBitrate with which the header is send */
const double HEADER_BITRATE_A=6E+6;
const double HEADER_BITRATE_B=1E+6;

// time slot ST, short interframe space SIFS, distributed interframe
// space DIFS, and extended interframe space EIFS
const simtime_t ST_A = 9E-6;
const simtime_t ST_B = 20E-6;
const simtime_t SIFS_A = 16E-6;
const simtime_t SIFS_B = 10E-6;
const simtime_t DIFS_A = 2*ST_A + SIFS_A;
const simtime_t DIFS_B = 2*ST_B + SIFS_B;
const simtime_t EIFS_A = SIFS_A + DIFS_A + PLCP_PREAMBLE_DELAY + (PHY_HEADER_LENGTH_A + LENGTH_ACK)/HEADER_BITRATE_A;
const simtime_t EIFS_B = SIFS_B + DIFS_B + PLCP_PREAMBLE_DELAY + (PHY_HEADER_LENGTH_B + LENGTH_ACK)/HEADER_BITRATE_B;
const simtime_t MAX_PROPAGATION_DELAY = 2E-6;  // 300 meters at the speed of light

#endif

