//
// Copyright (c) 2007 INRIA
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2 as
// published by the Free Software Foundation;
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// @author Ahmed Ayadi < Ahmed.Ayadi@sophia.inria.fr>
// @date 2007-10-11
//
#include "LbmsMacAP.h"
#include "RadioState.h"
#include "InterfaceTable.h"
#include "InterfaceTableAccess.h"
#include "PhyControlInfo_m.h"

Define_Module(LbmsMacAP);


/****************************************************************
 * Initialization functions.
 */
void LbmsMacAP::initialize(int stage)
{
    Ieee80211Mac::initialize(stage);

    if (stage == 0)
    {
        EV << "Initializing stage 0\n";
        // state variables

		const char * choice = par("leaderChoice");

        if (!strcmp(choice, "Worst"))
            leaderChoice = Worst;
        else if (!strcmp(choice, "Medium"))
            leaderChoice = Medium;
		else
            leaderChoice = Best;

        const char * leader = par("leaderElection");
        if (!strcmp(leader, "SNR"))
			leaderElection = SNR;
        else
			leaderElection = PACKET_LOST_RATE;
        fsm.setName("LbmsMacAP State Machine");
        // statistics
        numReceivedMulticast = 0;

        WATCH(numReceivedMulticast);
    }
}


/****************************************************************
 * Message handling functions.
 */

void LbmsMacAP::handleUpperMsg(cMessage *msg)
{
    EV<<"Ieee80211Mac handle message from upper layer"<<endl;
    // check if it's a command from the mgmt layer
    if (msg->length()==0 && msg->kind()!=0)
    {
        handleCommand(msg);
        return;
    }

    // check for queue overflow
    if (maxQueueSize && transmissionQueue.size() == maxQueueSize)
    {
        EV << "message " << msg << " received from higher layer but MAC queue is full, dropping message\n";
        delete msg;
        return;
    }

    if (!strcmp(msg->name(), "LBMS_report"))
    {
        ev<<"::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::"<< endl;
        ev<<parentModule()->fullPath()<<"."<<className()<<":"<<"LBMSReportFrame"<< endl;
        EV<<"Lbms Report Frame"<<endl;
        Ieee80211LBMSReport *lbmsReport = check_and_cast<Ieee80211LBMSReport *>(msg);

        for (MulticastGroupListI j = MulticastGrouplist.begin(); j != MulticastGrouplist.end(); j++)
        {
            int bestIndex = 0;
            int lastLeaderIndex = 0;
            if (leaderElection == PACKET_LOST_RATE)
            {
                for (int i=0; i<j->MAClist.size(); i++)
                {
                     if (j->MAClist.at(i).packetloss > j->MAClist.at(bestIndex).packetloss)
                         bestIndex = i;
                     if (j->MAClist.at(i).isLeader)
                         lastLeaderIndex = i;
                }
            }
            else
            {
                for (int i=0; i<j->MAClist.size(); i++)
                {
                     if (j->MAClist.at(i).snr < j->MAClist.at(bestIndex).snr)
                         bestIndex = i;
                     if (j->MAClist.at(i).isLeader)
                         lastLeaderIndex = i;
                }
            }

            if (! j->MAClist.at(bestIndex).isLeader)
            {
                j->MAClist.at(bestIndex).isLeader=true;
                j->MAClist.at(lastLeaderIndex).isLeader=false;
                EV<<": New Election "<<endl;
                EV<<": Last Leader " << j->MAClist.at(lastLeaderIndex).address<<endl;
                EV<<": New Leader "<< j->MAClist.at(bestIndex).address<< endl;

                ev<<parentModule()->fullPath()<<"."<<className()<<":New Election " << j->MAClist.at(lastLeaderIndex).address<<endl;
                ev<<parentModule()->fullPath()<<"."<<className()<<":Last Leader " << j->MAClist.at(lastLeaderIndex).address<<endl;
                ev<<parentModule()->fullPath()<<"."<<className()<<":New Leader "<< j->MAClist.at(bestIndex).address<< endl;

                Ieee80211LBMSReport * lbmsReport2 = new Ieee80211LBMSReport("LBMS_report");
                lbmsReport2->setReceiverAddress(j->MAClist.at(lastLeaderIndex).address);
                lbmsReport2->setElementArraySize(1);
                lbmsReport2->setAddress3(address);
                lbmsReport2->getElement(0).multicastGroupAddress = j->groupAddress;
                lbmsReport2->getElement(0).leader = false;
                lbmsReport2->setTransmitterAddress(address);

                lbmsReport2->setSequenceNumber(sequenceNumber);
                sequenceNumber = (sequenceNumber+1) % 4096;
                transmissionQueue.push_back(lbmsReport2);
                handleWithFSM(lbmsReport2);

                lbmsReport->setElementArraySize(1);
                lbmsReport->setReceiverAddress(j->MAClist.at(bestIndex).address);
                lbmsReport->setAddress3(address);
                lbmsReport->getElement(0).multicastGroupAddress = j->groupAddress;
                lbmsReport->getElement(0).leader = true;
                lbmsReport->setTransmitterAddress(address);

            }
        }

        lbmsReport->setSequenceNumber(sequenceNumber);
        sequenceNumber = (sequenceNumber+1) % 4096;
        transmissionQueue.push_back(lbmsReport);

        handleWithFSM(lbmsReport);
        return;
        ev<<"::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::"<< endl;

    }

    // must be a Ieee80211DataOrMgmtFrame, within the max size because we don't support fragmentation
    Ieee80211DataOrMgmtFrame *frame = check_and_cast<Ieee80211DataOrMgmtFrame *>(msg);
    if (frame->byteLength() > fragmentationThreshold)
        error("message from higher layer (%s)%s is too long for 802.11b, %d bytes (fragmentation is not supported yet)",
              msg->className(), msg->name(), msg->byteLength());
    EV << "frame " << frame << " received from higher layer, receiver = " << frame->getReceiverAddress() << endl;

    ASSERT(!frame->getReceiverAddress().isUnspecified());

    // fill in missing fields (receiver address, seq number), and insert into the queue
    frame->setTransmitterAddress(address);
    frame->setSequenceNumber(sequenceNumber);
    sequenceNumber = (sequenceNumber+1) % 4096;  //XXX seqNum must be checked upon reception of frames!

    transmissionQueue.push_back(frame);

    handleWithFSM(frame);
}

/**
 * Msg can be upper, lower, self or NULL (when radio state changes)
 */

void LbmsMacAP::handleWithFSM(cMessage *msg)
{
    // skip those cases where there's nothing to do, so the switch looks simpler
    if (isUpperMsg(msg) && fsm.state() != IDLE)
    {
        EV << "deferring upper message transmission in " << fsm.stateName() << " state\n";
        return;
    }

    Ieee80211Frame *frame = dynamic_cast<Ieee80211Frame*>(msg);

    int frameType = frame ? frame->getType() : -1;
    int msgKind = msg->kind();
    logState();
    //stateVector.record(fsm.state());

    if (frame && isLowerMsg(frame))
    {
        lastReceiveFailed =(msgKind == COLLISION || msgKind == BITERROR);
        scheduleReservePeriod(frame);
    }

    FSMA_Switch(fsm)
    {
        FSMA_State(IDLE)
        {
            FSMA_Enter(sendDownPendingRadioConfigMsg());
            FSMA_Event_Transition(Data-Ready,
                                  isUpperMsg(msg),
                                  DEFER,
                invalidateBackoffPeriod();
            );
            FSMA_No_Event_Transition(Immediate-Data-Ready,
                                     !transmissionQueue.empty(),
                                     DEFER,
                invalidateBackoffPeriod();
            );
            FSMA_Event_Transition(Receive,
                                  isLowerMsg(msg),
                                  RECEIVE,
            );
        }
        FSMA_State(DEFER)
        {
            FSMA_Enter(sendDownPendingRadioConfigMsg());
            FSMA_Event_Transition(Wait-DIFS,
                                  isMediumStateChange(msg) && isMediumFree(),
                                  WAITDIFS,
            ;);
            FSMA_No_Event_Transition(Immediate-Wait-DIFS,
                                     isMediumFree() || !backoff,
                                     WAITDIFS,
            ;);
            FSMA_Event_Transition(Receive,
                                  isLowerMsg(msg),
                                  RECEIVE,
            ;);
        }
        FSMA_State(WAITDIFS)
        {
            FSMA_Enter(scheduleDIFSPeriod());
            FSMA_Event_Transition(Immediate-Transmit-RTS,
                                  msg == endDIFS && !isBroadcast(currentTransmission()) && !isMulticast(currentTransmission())
                                  && currentTransmission()->byteLength() >= rtsThreshold && !backoff,
                                  WAITCTS,
                sendRTSFrame(currentTransmission());
                cancelDIFSPeriod();
            );
            FSMA_Event_Transition(Immediate-Transmit-Broadcast,
                                  msg == endDIFS && isBroadcast(currentTransmission()) && !backoff,
                                  WAITBROADCAST,
                sendBroadcastFrame(currentTransmission());
                cancelDIFSPeriod();
            );
            FSMA_Event_Transition(Immediate-Transmit-Multicast,
                                  msg == endDIFS && isMulticast(currentTransmission()) 
								  && !isBroadcast(currentTransmission()) && !backoff
								  && isRegistredGroup(currentTransmission()),
                                  WAITACK,
                sendMulticastFrame(currentTransmission());
                cancelDIFSPeriod();
            );
            FSMA_Event_Transition(Immediate-Transmit-Data,
                                  msg == endDIFS && !isBroadcast(currentTransmission()) 
								  && !isMulticast(currentTransmission()) && !backoff,
                                  WAITACK,
                sendDataFrame(currentTransmission());
                cancelDIFSPeriod();
            );
            FSMA_Event_Transition(DIFS-Over,
                                  msg == endDIFS,
                                  BACKOFF,
                ASSERT(backoff);
                if (isInvalidBackoffPeriod())
                    generateBackoffPeriod();
            );
            FSMA_Event_Transition(Busy,
                                  isMediumStateChange(msg) && !isMediumFree(),
                                  DEFER,
                backoff = true;
                cancelDIFSPeriod();
            );
            FSMA_No_Event_Transition(Immediate-Busy,
                                     !isMediumFree(),
                                     DEFER,
                backoff = true;
                cancelDIFSPeriod();
            );
            // radio state changes before we actually get the message, so this must be here
            FSMA_Event_Transition(Receive,
                                  isLowerMsg(msg),
                                  RECEIVE,
                cancelDIFSPeriod();
            ;);
        }
        FSMA_State(BACKOFF)
        {
            FSMA_Enter(scheduleBackoffPeriod());
            FSMA_Event_Transition(Transmit-RTS,
                                  msg == endBackoff && !isBroadcast(currentTransmission()) && !isMulticast(currentTransmission())
                                  && currentTransmission()->byteLength() >= rtsThreshold,
                                  WAITCTS,
                sendRTSFrame(currentTransmission());
            );//ahmed
            FSMA_Event_Transition(Transmit-Broadcast,
                                  msg == endBackoff && isBroadcast(currentTransmission()),
                                  WAITBROADCAST,
                sendBroadcastFrame(currentTransmission());
            );
            FSMA_Event_Transition(Transmit-Multicast,
                                  msg == endBackoff && isMulticast(currentTransmission()),
                                  WAITACK,
                sendMulticastFrame(currentTransmission());
            );//ahmed
            FSMA_Event_Transition(Transmit-Data,
                                  msg == endBackoff && !isBroadcast(currentTransmission()) && !isMulticast(currentTransmission()),
                                  WAITACK,
                sendDataFrame(currentTransmission());
            );//ahmed
            FSMA_Event_Transition(Backoff-Busy,
                                  isMediumStateChange(msg) && !isMediumFree(),
                                  DEFER,
                cancelBackoffPeriod();
                decreaseBackoffPeriod();
            );
        }
        FSMA_State(WAITACK)
        {
            FSMA_Enter(scheduleDataTimeoutPeriod(currentTransmission()));
            FSMA_Event_Transition(Receive-ACK,
                                  isLowerMsg(msg) && isForUs(frame) && frameType == ST_ACK,
                                  IDLE,
                cancelTimeoutPeriod();
                finishCurrentTransmission();
                if (retryCounter == 0) numSentWithoutRetry++;
                numSent++;
            );
            FSMA_Event_Transition(Transmit-Data-Failed,
                                  msg == endTimeout && retryCounter == retryLimit,
                                  IDLE,
                giveUpCurrentTransmission();
            );
            FSMA_Event_Transition(Receive-ACK-Timeout,
                                  msg == endTimeout,
                                  DEFER,
                retryCurrentTransmission();
            );
        }
        // wait until broadcast is sent
        FSMA_State(WAITBROADCAST)
        {
            FSMA_Enter(scheduleBroadcastTimeoutPeriod(currentTransmission()));
            FSMA_Event_Transition(Transmit-Broadcast,
                                  msg == endTimeout,
                                  IDLE,
                finishCurrentTransmission();
                numSentBroadcast++;
            );
        }

        FSMA_State(WAITMULTICAST)
        {
            FSMA_Enter(scheduleMulticastTimeoutPeriod(currentTransmission()));
            FSMA_Event_Transition(Transmit-Multicast,
                                  msg == endTimeout,
                                  IDLE,
                finishCurrentTransmission();
            );
        }//ahmed
        // accoriding to 9.2.5.7 CTS procedure
        FSMA_State(WAITCTS)
        {
            FSMA_Enter(scheduleCTSTimeoutPeriod());
            FSMA_Event_Transition(Receive-CTS,
                                  isLowerMsg(msg) && isForUs(frame) && frameType == ST_CTS,
                                  WAITSIFS,
                cancelTimeoutPeriod();
            );
            FSMA_Event_Transition(Transmit-RTS-Failed,
                                  msg == endTimeout && retryCounter == retryLimit,
                                  IDLE,
                giveUpCurrentTransmission();
            );
            FSMA_Event_Transition(Receive-CTS-Timeout,
                                  msg == endTimeout,
                                  DEFER,
                retryCurrentTransmission();
            );
        }
        FSMA_State(WAITSIFS)
        {
            FSMA_Enter(scheduleSIFSPeriod(frame));
            FSMA_Event_Transition(Transmit-CTS,
                                  msg == endSIFS && frameReceivedBeforeSIFS()->getType() == ST_RTS,
                                  IDLE,
                sendCTSFrameOnEndSIFS();
                resetStateVariables();
            );
            FSMA_Event_Transition(Transmit-DATA,
                                  msg == endSIFS && frameReceivedBeforeSIFS()->getType() == ST_CTS,
                                  WAITACK,
                sendDataFrameOnEndSIFS(currentTransmission());
            );
            FSMA_Event_Transition(Transmit-ACK,
                                  msg == endSIFS && isDataOrMgmtFrame(frameReceivedBeforeSIFS()),
                                  IDLE,
                sendACKFrameOnEndSIFS();
                resetStateVariables();
            );
        }
        // this is not a real state
        FSMA_State(RECEIVE)
        {
            FSMA_No_Event_Transition(Immediate-Receive-Error,
                                     isLowerMsg(msg) && (msgKind == COLLISION || msgKind == BITERROR),
                                     IDLE,
                EV << "received frame contains bit errors or collision, next wait period is EIFS\n";
                numCollision++;
                resetStateVariables();
            );

            FSMA_No_Event_Transition(Immediate-Receive-LBMS,
                                     isLowerMsg(msg) && (frameType == ST_LBMS_REPORT || frameType == ST_LBMS_REQUEST),
                                     WAITSIFS,
                 processIncomingLBMS(frame);
            );

            FSMA_No_Event_Transition(Immediate-Receive-Broadcast,
                                     isLowerMsg(msg) && isBroadcast(frame) && isDataOrMgmtFrame(frame),
                                     IDLE,
                sendUp(frame);
                numReceivedBroadcast++;
                resetStateVariables();
            );
            FSMA_No_Event_Transition(Immediate-Receive-Multicast,
                                     isLowerMsg(msg) && isMulticast(frame) && isDataOrMgmtFrame(frame),
                                     IDLE,
                processIncomingMulticastFrame(frame);
                //sendUp(frame);
            );

            FSMA_No_Event_Transition(Immediate-Receive-Data,
                                     isLowerMsg(msg) && isForUs(frame) && isDataOrMgmtFrame(frame),
                                     WAITSIFS,
                sendUp(frame);
                numReceived++;
            );

            FSMA_No_Event_Transition(Immediate-Receive-RTS,
                                     isLowerMsg(msg) && isForUs(frame) && frameType == ST_RTS,
                                     WAITSIFS,
            );
            FSMA_No_Event_Transition(Immediate-Receive-Other,
                                     isLowerMsg(msg),
                                     IDLE,
                resetStateVariables();
            );
        }
    }

    logState();
    //stateVector.record(fsm.state());
}


/****************************************************************
 * Timer functions.
 */

void LbmsMacAP::scheduleMulticastTimeoutPeriod(Ieee80211DataOrMgmtFrame *frameToSend)
{
    EV << "scheduling multicast timeout period\n";
    scheduleAt(simTime() + frameDuration(frameToSend), endTimeout);
}

/****************************************************************
 * Frame sender functions.
 */


void LbmsMacAP::sendMulticastFrame(Ieee80211DataOrMgmtFrame *frameToSend)
{
    sendDown(buildMulticastFrame(frameToSend));
}


Ieee80211DataOrMgmtFrame *LbmsMacAP::buildMulticastFrame(Ieee80211DataOrMgmtFrame *frameToSend)
{
    Ieee80211DataOrMgmtFrame *frame = (Ieee80211DataOrMgmtFrame *)frameToSend->dup();
    frame->setDuration(SIFSPeriod() + frameDuration(LENGTH_ACK, basicBitrate));
    return frame;
}

/****************************************************************
 * Helper functions.
 */

bool LbmsMacAP::isMulticast(Ieee80211Frame *frame)
{	
	if (frame && (frame->getReceiverAddress().isMulticast()))
		return true;
	else 
	{
		Ieee80211DataOrMgmtFrame *mframe = (Ieee80211DataOrMgmtFrame *) frame;
		if (mframe && (mframe->getAddress3().isMulticast()))
			return true;
		else
			return false;
	}
}

bool LbmsMacAP::isForUs(Ieee80211Frame *frame)
{
    return frame && frame->getReceiverAddress() == address;
}


bool LbmsMacAP::isNewGroup(Ieee80211Frame *frame)
{
//	std::cout<<"isNewGroup"<<endl;
    //Ieee80211LBMSRequest *mframe = check_and_cast<Ieee80211LBMSRequest *>(frame);
    //ev<<parentModule()->fullPath()<<"."<<className()<<":" <<"Transmitter Address :"<<mframe->getTransmitterAddress()<< endl;
    //ev<<parentModule()->fullPath()<<"."<<className()<<":" <<"Multicast Address :"<<mframe->getElement().multicastAddress<< endl;

	Ieee80211LBMSRequest *reqframe;
	Ieee80211LBMSReport *repframe;
    if (reqframe = dynamic_cast<Ieee80211LBMSRequest *>(frame))
	{
//		std::cout<< " Ieee80211LBMSRequest " << endl;
		if (MulticastGrouplist.empty())
		{
//			std::cout<<parentModule()->fullPath()<<"."<<className()<<":" <<"isNewGroup true"<< endl;
			return true;
		}
		else 
		{
			for (MulticastGroupListI j = MulticastGrouplist.begin (); j != MulticastGrouplist.end (); j++)
			{
				if (j->groupAddress == reqframe->getElement().multicastAddress)
				{
					return false;
//					std::cout <<parentModule()->fullPath()<<"."<<className()<<":" <<"isNewGroup false"<< endl;
				}
			}
		}
//		std::cout <<parentModule()->fullPath()<<"."<<className()<<":" <<"isNewGroup true"<< endl;
		return true;
	}
	else
	{
//		std::cout<< " Ieee80211LBMSReport " << endl;
		repframe = dynamic_cast<Ieee80211LBMSReport *>(frame);
		if (MulticastGrouplist.empty())
		{
//			std::cout<<parentModule()->fullPath()<<"."<<className()<<":" <<"isNewGroup true"<< endl;
			return true;
		}
		else 
		{
			for (MulticastGroupListI j = MulticastGrouplist.begin (); j != MulticastGrouplist.end (); j++)
			{
				if (j->groupAddress == repframe->getElement(0).multicastGroupAddress)
				{
					return false;
//					std::cout<<parentModule()->fullPath()<<"."<<className()<<":" <<"isNewGroup false"<< endl;
				}
			}
		}
//		std::cout<<parentModule()->fullPath()<<"."<<className()<<":" <<"isNewGroup true"<< endl;
		return true;
	}
}


bool LbmsMacAP::isRegistredGroup(Ieee80211DataOrMgmtFrame *frame)
{
    if (MulticastGrouplist.empty())
    {
         return false;
    }
    else 
    {
        for (MulticastGroupListI j = MulticastGrouplist.begin (); j != MulticastGrouplist.end (); j++)
        {
            ev<<parentModule()->fullPath()<<"."<<className()<<":"<< "Comparer "<< j->groupAddress<< " & " << frame->getReceiverAddress() <<endl; 
            if (j->groupAddress == frame->getReceiverAddress())
            {
                ev<<parentModule()->fullPath()<<"."<<className()<<":"<<"isRegistredGroup true"<< endl;
                return true;
            }
        }
    }
    ev<<parentModule()->fullPath()<<"."<<className()<<":"<<"isRegistredGroup false"<< endl;
    return false;
}


bool LbmsMacAP::isNewMember(Ieee80211Frame *frame)
{
	//	std::cout<<"isNewMember"<<endl;
    //Ieee80211LBMSRequest *mframe = check_and_cast<Ieee80211LBMSRequest *>(frame);
    //std::cout<<parentModule()->fullPath()<<"."<<className()<<":" <<"Transmitter Address :"<<mframe->getTransmitterAddress()<< endl;
    //std::cout<<parentModule()->fullPath()<<"."<<className()<<":" <<"Multicast Address :"<<mframe->getElement().multicastAddress<< endl;

	Ieee80211LBMSRequest *reqframe;
	Ieee80211LBMSReport *repframe;
    if (reqframe = dynamic_cast<Ieee80211LBMSRequest *>(frame))
	{
		for (MulticastGroupListI j = MulticastGrouplist.begin (); j != MulticastGrouplist.end (); j++)
		{
//         std::cout<< "Compare "<< j->groupAddress<< " & " << reqframe->getElement().multicastAddress <<endl;
         if (j->groupAddress == reqframe->getElement().multicastAddress)
         {
             for (MulticastMACAddressListI i = j->MAClist.begin (); i != j->MAClist.end (); i++)
             {
                 //std::cout<< "Compare "<</* */i->address<< " & " << reqframe->getTransmitterAddress() <<endl; 
                 if (i->address == reqframe->getTransmitterAddress())
                 {
//                      std::cout<< "loop - Member Address false" <<i->address<<endl;
                      return false ;
                 }
             }
         }
		}
//		std::cout<<"isNewMember true"<<endl;
		return true;
	}
	else
	{
		repframe = check_and_cast<Ieee80211LBMSReport *>(frame);
		for (MulticastGroupListI j = MulticastGrouplist.begin (); j != MulticastGrouplist.end (); j++)
		{
//         std::cout<< "Compare "<< j->groupAddress<< " & " << repframe->getElement(0).multicastGroupAddress <<endl;
         if (j->groupAddress == repframe->getElement(0).multicastGroupAddress)
         {
             for (MulticastMACAddressListI i = j->MAClist.begin (); i != j->MAClist.end (); i++)
             {
              //   std::cout<< "Compare "<</* */i->address<< " & " << repframe->getTransmitterAddress() <<endl; 
                 if (i->address == repframe->getTransmitterAddress())
                 {
//                      std::cout<< "loop - Member Address false" <<i->address<<endl;
                      return false ;
                 }
             }
         }
		}
//		std::cout<<"isNewMember true"<<endl;
		return true;
	}
}


bool LbmsMacAP::isAssociatedGroup(Ieee80211DataOrMgmtFrame *frameToSend)
{
    ev<<parentModule()->fullPath()<<"."<<className()<<":"<<"isAssociatedGroup\n";
    myMulticastGroup();
    ev<<"getTransmitterAddress  = "<<frameToSend->getTransmitterAddress()<<endl;
    ev<<"getReceiverAddres  = "<<frameToSend->getReceiverAddress()<<endl;
    ev<<"getAddress3  = "<<frameToSend->getAddress3();
    for (MulticastGroupListI j = MulticastGrouplist.begin (); j != MulticastGrouplist.end (); j++)
	{
        ev<<"Compare "<<j->groupAddress<<" & "<<frameToSend->getAddress3()<<endl;
        if (j->groupAddress == frameToSend->getAddress3()) 
        {
            ev<<" true"<<endl;
            return true;
        }
    }
    ev<<" false"<<endl;
    return false;
}


void LbmsMacAP::setLeader(Ieee80211DataOrMgmtFrame *frameToSend)
{
    for (MulticastGroupListI j = MulticastGrouplist.begin (); j != MulticastGrouplist.end (); j++) {
         if (j->groupAddress == frameToSend->getTransmitterAddress())
         {
              for (MulticastMACAddressListI i = j->MAClist.begin (); i != j->MAClist.end (); j++)
              if (i->address == frameToSend->getReceiverAddress()) i->isLeader = true;
         }
    }
}


void LbmsMacAP::setNonLeader(Ieee80211DataOrMgmtFrame *frameToSend)
{
    for (MulticastGroupListI j = MulticastGrouplist.begin (); j != MulticastGrouplist.end (); j++) {
         if (j->groupAddress == frameToSend->getReceiverAddress())
         {
              for (MulticastMACAddressListI i = j->MAClist.begin (); i != j->MAClist.end (); j++)
              if (i->address == frameToSend->getTransmitterAddress()) i->isLeader = false;
         }
    }
}


void LbmsMacAP::processIncomingLBMS(Ieee80211Frame *frame)
{
    if (frame->getType() == ST_LBMS_REQUEST)
    {
        processLBMSRequestFrame(frame);
    }
    else if (frame->getType() == ST_LBMS_REPORT)
        processLBMSReportFrame(frame);
    else delete frame;
}


void LbmsMacAP::processIncomingMulticastFrame(Ieee80211Frame *frame)
{
    ev<<"processIncomingMulticastFrame "<<frame->name()<<endl;
    Ieee80211DataOrMgmtFrame *mframe = (Ieee80211DataOrMgmtFrame *) frame;
    if (isAssociatedGroup(mframe)) sendUp(frame);
}


void LbmsMacAP::addNewMulticastGroup(Ieee80211Frame *frame)
{
	MulticastMACAddress macAddress;
	MulticastGroup multicastGroup;
	
	Ieee80211LBMSRequest *reqframe;
	Ieee80211LBMSReport *repframe;

    if (reqframe = dynamic_cast<Ieee80211LBMSRequest *>(frame))
	{
		ev<< "NewMulticastGroup "<<reqframe->getElement().multicastAddress<<endl;
		ev<< "A New Group Member "<<reqframe->getTransmitterAddress()<<endl;
		macAddress.address = reqframe->getTransmitterAddress();
		macAddress.isLeader = true;
		macAddress.packetloss = 0;
		multicastGroup.groupAddress = reqframe->getElement().multicastAddress;

	}
	else
	{
		repframe = dynamic_cast<Ieee80211LBMSReport *>(frame);
		ev<< "NewMulticastGroup "<<repframe->getElement(0).multicastGroupAddress<<endl;
		ev<< "A New Group Member "<<repframe->getTransmitterAddress()<<endl;
		macAddress.address = repframe->getTransmitterAddress();
		macAddress.isLeader = true;
		macAddress.packetloss = 0;
		multicastGroup.groupAddress = repframe->getElement(0).multicastGroupAddress;
	}
	multicastGroup.MAClist.push_back(macAddress);
	MulticastGrouplist.push_back(multicastGroup);
	myMulticastGroup();
}


void LbmsMacAP::addNewMulticastMember(Ieee80211Frame *frame)
{
    Ieee80211LBMSRequest *reqframe;
	Ieee80211LBMSReport *repframe;
	MulticastMACAddress macAddress;
	if (reqframe= dynamic_cast<Ieee80211LBMSRequest *>(frame))
	{
		ev<< "A New Group Member "<<reqframe->getTransmitterAddress()<<endl;
		ev<< "Group Address "<<reqframe->getElement().multicastAddress<<endl;
		macAddress.address = reqframe->getTransmitterAddress();
		macAddress.isLeader = false;
		macAddress.packetloss = 0;
		macAddress.snr = 0;
		for (MulticastGroupListI j = MulticastGrouplist.begin (); j != MulticastGrouplist.end (); j++)
		{
			if (j->groupAddress ==  reqframe->getElement().multicastAddress)
			j->MAClist.push_back(macAddress);
		}
	}
	else
	{
		repframe= dynamic_cast<Ieee80211LBMSReport *>(frame);
		ev<< "A New Group Member "<<repframe->getTransmitterAddress()<<endl;
		ev<< "Group Address "<<repframe->getElement(0).multicastGroupAddress<<endl;
		macAddress.address = repframe->getTransmitterAddress();
		macAddress.isLeader = false;
		macAddress.packetloss = 0;
		macAddress.snr = 0;
		for (MulticastGroupListI j = MulticastGrouplist.begin (); j != MulticastGrouplist.end (); j++)
		{
			if (j->groupAddress ==  repframe->getElement(0).multicastGroupAddress)
			j->MAClist.push_back(macAddress);
		}
	}
    myMulticastGroup();
}


void LbmsMacAP::RemoveMulticastMember(Ieee80211DataOrMgmtFrame *frame)
{
    ev<<parentModule()->fullPath()<<"."<<className()<<":"<<"Remove Multicast Group Member : ";
    Ieee80211LBMSRequest *mframe = check_and_cast<Ieee80211LBMSRequest *>(frame);
    ev << mframe->getTransmitterAddress() <<endl;
    ev << "Multicast Address "<<mframe->getElement().multicastAddress <<endl;
    bool isLeader = true;

    for (MulticastGroupListI j = MulticastGrouplist.begin(); j != MulticastGrouplist.end(); j++)
    {
         ev << "Compare "<< j->groupAddress<< " & " << mframe->getElement().multicastAddress <<endl;
         if (j->groupAddress == mframe->getElement().multicastAddress )
         {
             for (MulticastMACAddressListI i = j->MAClist.begin (); i != j->MAClist.end (); i++)
             {
                 ev<< "Compare "<< i->address<< " & " << mframe->getTransmitterAddress() <<endl;
                 if (i->address == mframe->getTransmitterAddress())
                 {
                     isLeader = i->isLeader;
                     j->MAClist.erase(i);
                     break;
                 }
             }
         }
         if (j->MAClist.empty()) 
         {
              MulticastGrouplist.erase(j);
              ev<<"Remove Multicast Group"<<mframe->getElement().multicastAddress<<endl;
         }
         else
         {
              if (isLeader) electNewLeader(mframe);
         }
         break;
    }
    myMulticastGroup();
}


void LbmsMacAP::myMulticastGroup()
{
    int k=1;
    int l=1;
	std::cout<<parentModule()->fullPath()<<"."<<className()<<": Log Multicast Group" <<endl;
    std::cout<< "------------------------------------------------------------------" <<endl;
    for (MulticastGroupListI j = MulticastGrouplist.begin (); j != MulticastGrouplist.end (); j++) {
         std::cout<<"Multicast Group "<< k++ <<" Multicast Address : " <<j->groupAddress <<endl;
         std::cout<<"Multicast Member Adress : "<<endl;
         l=1;
         for (MulticastMACAddressListI i = j->MAClist.begin (); i != j->MAClist.end (); i++)
         {
             std::cout<<"Membre "<< l++ <<" MACAddress:" <<i->address;
             std::cout<<" Packet loss "<<i->packetloss;
             std::cout<<" Snr "<<i->snr;
			 if (i->isLeader) std::cout<< " is Leader";
             std::cout<< endl;
         }
    }
    std::cout<< "------------------------Log Multicast Group END-------------------" <<endl;
}


void LbmsMacAP::processLBMSRequestFrame(Ieee80211Frame *mframe)
{
    Ieee80211DataOrMgmtFrame *frame = (Ieee80211DataOrMgmtFrame *) mframe;
    ev<<parentModule()->fullPath()<<"."<<className()<<":"<<"processIncomingLBMS Request"<<endl;
    ev<<"Received message  "<<frame->name()<<endl;
    ev<<"getTransmitterAddress :"<<frame->getTransmitterAddress()<< endl;

    if (!strcmp(frame->name(),"Join"))
    {
        if (isNewGroup(frame)) addNewMulticastGroup (frame);
        if (isNewMember(frame))
        {
            addNewMulticastMember(frame);
            return;
        }
    }
    else if (!strcmp(frame->name(),"Leave"))
    {
        RemoveMulticastMember(frame);
    }
    sendUp(frame);
}


void LbmsMacAP::processLBMSReportFrame(Ieee80211Frame *mframe)
{
    Ieee80211LBMSReport *frame = (Ieee80211LBMSReport *) mframe;
    ev<<parentModule()->fullPath()<<"."<<className()<<":"<<"processIncomingLBMS Report"<<endl;
    ev<<"getTransmitterAddress :"<<frame->getTransmitterAddress()<< endl;
    ev<<"Multicast Address :"<<frame->getElement(0).multicastGroupAddress<< endl;
    ev<<"Packet Loss Power = "<<frame->getPacketLoss()<<endl;

	/*if (isNewGroup(frame)) addNewMulticastGroup (frame);
    if (isNewMember(frame))
    {
        addNewMulticastMember(frame);
        return;
    }*/
    for (MulticastGroupListI j = MulticastGrouplist.begin(); j != MulticastGrouplist.end(); j++)
    {
         if (j->groupAddress == frame->getElement(0).multicastGroupAddress )
         {
             for (MulticastMACAddressListI i = j->MAClist.begin (); i != j->MAClist.end (); i++)
             {
                 if (i->address == frame->getTransmitterAddress())
                 {
                     i->packetloss = frame->getPacketLoss();
                     i->snr = frame->getSnr();
                     break;
                 }
             }
             break;
         }
    }
    myMulticastGroup();
    sendUp(frame);
}


void LbmsMacAP::electNewLeader(Ieee80211LBMSRequest *mframe)
{
    for (MulticastGroupListI j = MulticastGrouplist.begin(); j != MulticastGrouplist.end(); j++)
    {
         if (j->groupAddress == mframe->getElement().multicastAddress )
         {
             // choose best Leader
             int leaderIndex = chooseLeader(mframe);
             j->MAClist.at(leaderIndex).isLeader = true;
             Ieee80211LBMSRequest *frame = new Ieee80211LBMSRequest ("MACJoin");
             frame->setTransmitterAddress(j->MAClist.at(leaderIndex).address);
             frame->getElement().multicastAddress = mframe->getElement().multicastAddress;
             sendUp(frame);
         }
    }
}


int LbmsMacAP::chooseLeader(Ieee80211LBMSRequest *mframe)
{
    // here, just choose the one with the greatest receive power
    // TODO and which supports a good data rate we support
    int bestIndex = 0;
    for (MulticastGroupListI j = MulticastGrouplist.begin(); j != MulticastGrouplist.end(); j++)
    {
         EV << "Compare "<< j->groupAddress<< " & " << mframe->getElement().multicastAddress <<endl;
         if (j->groupAddress == mframe->getElement().multicastAddress )
         {
             // choose best Leader
            if (leaderElection == PACKET_LOST_RATE)
            {
				double meanLoss = 0;
				for (int i=0; i<j->MAClist.size(); i++)
                {
					if ((leaderChoice == Worst) && (j->MAClist.at(i).packetloss > j->MAClist.at(bestIndex).packetloss))
                        bestIndex = i;
					else if ((leaderChoice == Best) && (j->MAClist.at(i).packetloss < j->MAClist.at(bestIndex).packetloss))
                        bestIndex = i;
					else
						meanLoss += j->MAClist.at(i).packetloss/j->MAClist.size();
                }
				if (leaderChoice == Medium)
				{
					double differ = abs(j->MAClist.at(0).packetloss - meanLoss);
					bestIndex = 0;
					for (int i=1; i<j->MAClist.size(); i++)
					{
						if (abs(j->MAClist.at(i).packetloss - meanLoss) < differ)
						{
							bestIndex = i;
							differ = abs(j->MAClist.at(i).packetloss - meanLoss);
						}
					}
				}
            }
            else
            {
				double meanSNR = 0;
                for (int i=0; i<j->MAClist.size(); i++)
                {
                    if ((leaderChoice == Worst)&&(j->MAClist.at(i).snr < j->MAClist.at(bestIndex).snr))
                         bestIndex = i;
					else if ((leaderChoice == Best)&&(j->MAClist.at(i).snr > j->MAClist.at(bestIndex).snr))
                         bestIndex = i;
					else
						meanSNR += j->MAClist.at(i).snr/j->MAClist.size();
                }
				if (leaderChoice == Medium)
				{
					double differ = abs(j->MAClist.at(0).snr - meanSNR);
					bestIndex = 0;
					for (int i=1; i<j->MAClist.size(); i++)
					{
						if (abs(j->MAClist.at(i).snr - meanSNR) < differ)
						{
							bestIndex = i;
							differ = abs(j->MAClist.at(i).snr - meanSNR);
						}
					}
				}
             }
         }
         break;
    }
    return bestIndex;
}
void LbmsMacAP::sendLbmsReport(Ieee80211Frame *frame)
{
}

void LbmsMacAP::sendLbmsReport(bool leader, MACAddress &groupAddress, MACAddress &memberAddress)
{
}

