//
// Copyright (c) 2007 INRIA
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2 as
// published by the Free Software Foundation;
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// @author Ahmed Ayadi < Ahmed.Ayadi@sophia.inria.fr>
// @date 2007-10-11
//
#include "LbmsMgmtAP.h"
#include "Ieee802Ctrl_m.h"
#include "EtherFrame_m.h"
#include "NotifierConsts.h"
#include "RadioState.h"


Define_Module(LbmsMgmtAP);

static std::ostream& operator<< (std::ostream& os, const LbmsMgmtAP::STAInfo& sta)
{
    os << "state:" << sta.status;
    return os;
}

void LbmsMgmtAP::initialize(int stage)
{
    Ieee80211MgmtAP::initialize(stage);
    selectionInterval = par ("selectionInterval");
    numberOfParticipant = 0;
}

void LbmsMgmtAP::handleTimer(cMessage *msg)
{
    if (msg==beaconTimer)
    {
        sendBeacon();
        scheduleAt(simTime()+beaconInterval, beaconTimer);
    }
    else if (msg==lbmsTimer)
    {
        if (numberOfParticipant > 0) sendLbmsReport();
        scheduleAt(simTime()+selectionInterval, lbmsTimer);
    }
    else
    {
        error("internal error: unrecognized timer '%s'", msg->name());
    }
}

void LbmsMgmtAP::sendLbmsReport(Ieee80211LBMSRequest *frame)
{
    EV <<"Sending LBMS Report" << endl;
    EV<<"Sending LBMS Report" << endl;
    Ieee80211LBMSReport * lbmsReport = new Ieee80211LBMSReport("LBMS report");
    lbmsReport->setElementArraySize(1);
    lbmsReport->setFromDS(true);
    lbmsReport->setReceiverAddress(frame->getTransmitterAddress());
    lbmsReport->setAddress3(myAddress);
    lbmsReport->getElement(0).multicastGroupAddress = frame->getElement().multicastAddress;
    lbmsReport->getElement(0).leader = true;
    sendOrEnqueue(lbmsReport);
    delete frame;
}

void LbmsMgmtAP::sendLbmsReport()
{
    EV <<parentModule()->fullPath()<<"."<<className()<<":"<<"NumberofParticipant "<< numberOfParticipant<< endl;
    Ieee80211LBMSReport * lbmsReport = new Ieee80211LBMSReport("LBMS_report");
    lbmsReport->setAddress3(myAddress);
    lbmsReport->setElementArraySize(1);
    lbmsReport->getElement(0).leader = true;
    lbmsReport->setToDS(false);
    lbmsReport->setFromDS(false);
    sendOrEnqueue(lbmsReport);
    return;
}

void LbmsMgmtAP::handleLbmsRequestFrame(Ieee80211LBMSRequest *frame)
{
    if (!(strcmp(frame->name(),"Join") & strcmp(frame->name(),"MACJoin")))
    {
        EV <<parentModule()->fullPath()<<"."<<className()<<":"<<"HandleLBMSRequestFrame "<< frame->name()<<endl;
        if (numberOfParticipant == 0) 
        {
            lbmsTimer = new cMessage("lbmsTimer");
            scheduleAt(simTime() + selectionInterval, lbmsTimer);
        }
        if (!strcmp(frame->name(),"Join")) numberOfParticipant++;

        EV <<parentModule()->fullPath()<<"."<<className()<<":"<<"Number of participition "<< numberOfParticipant<< endl;
        if (numberOfParticipant < 2)  sendLbmsReport(frame);
        return;
    }
    else if (!strcmp(frame->name(),"Leave"))
    {
        EV <<parentModule()->fullPath()<<"."<<className()<<":"<<"HandleLBMSRequestFrame "<< frame->name()<<endl;
        numberOfParticipant--;
        EV <<parentModule()->fullPath()<<"."<<className()<<":"<<"Number of participition "<< numberOfParticipant<< endl;
    }
    delete frame;
}

void LbmsMgmtAP::handleLbmsReportFrame(Ieee80211LBMSReport *frame)
{
    EV <<parentModule()->fullPath()<<"."<<className()<<":"<<"HandleLBMSReportFrame"<<endl;
    delete frame;
}

void LbmsMgmtAP::handleDataFrame(Ieee80211DataFrame *frame)
{
    // check toDS bit
    if (!frame->getToDS())
    {
        // looks like this is not for us - discard
        EV << "frame->getToDS()" <<  frame->getToDS() <<endl;
        EV <<parentModule()->fullPath()<<"."<<className()<<" message deleted "<<endl;
        delete frame;
        return;
    }

    // look up destination address in our STA list
    EV << " AP frame->getAddress3() " << frame->getAddress3() <<endl;
    STAList::iterator it = staList.find(frame->getAddress3());
    if (it==staList.end())
    {
        // not our STA -- pass up frame to relayUnit for LAN bridging if we have one
        if (hasRelayUnit)
            send(convertToEtherFrame(frame), "uppergateOut");
        else
            distributeReceivedDataFrame(frame);
            //delete frame;
    }
    else
    {
        // dest address is our STA, but is it already associated?
        if (it->second.status == ASSOCIATED)
            distributeReceivedDataFrame(frame); // send it out to the destination STA
        else
        {
            delete frame;
        }
    }
}

