//
// Copyright (c) 2007 INRIA
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2 as
// published by the Free Software Foundation;
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// @author Ahmed Ayadi < Ahmed.Ayadi@sophia.inria.fr>
// @date 2007-10-11
//
#ifndef IEEE80211_MGMT_AP_LBMS_H
#define IEEE80211_MGMT_AP_LBMS_H

#include <omnetpp.h>
#include <map>
#include "Ieee80211MgmtAPBase.h"
#include "Ieee80211MgmtAP.h"
#include "NotificationBoard.h"


/**
 * Used in 802.11 infrastructure mode: handles management frames for
 * an access point (AP). See corresponding NED file for a detailed description.
 *
 * @author Andras Varga
 */
class INET_API LbmsMgmtAP : public Ieee80211MgmtAP
{
  public:
    /** State of a STA */
    enum STAStatus {NOT_AUTHENTICATED, AUTHENTICATED, ASSOCIATED};

    /** Describes a STA */
    struct STAInfo {
        MACAddress address;
        STAStatus status;
        int authSeqExpected;  // when NOT_AUTHENTICATED: transaction sequence number of next expected auth frame
        //int consecFailedTrans;  //XXX
        //double expiry;          //XXX association should expire after a while if STA is silent?
    };

    struct MAC_compare {
        bool operator()(const MACAddress& u1, const MACAddress& u2) const {return u1.compareTo(u2) < 0;}
    };
    typedef std::map<MACAddress,STAInfo, MAC_compare> STAList;

  protected:
    // configuration
    std::string ssid;
    int channelNumber;
    int selectionInterval;
    int numberOfParticipant;
    simtime_t lbmsInterval;
    int numAuthSteps;
    Ieee80211SupportedRatesElement supportedRates;

    // state
    STAList staList; ///< list of STAs
    cMessage *lbmsTimer;

  protected:
    virtual int numInitStages() const {return 2;}
    virtual void initialize(int);

    /** Implements abstract Ieee80211MgmtBase method */
    virtual void handleTimer(cMessage *msg);

    /** Utility function: creates and sends a lbms frame */
    void sendLbmsReport(Ieee80211LBMSRequest *frame);
    void sendLbmsReport();

    /** @name Processing of different frame types */
    //@{
    virtual void handleLbmsRequestFrame(Ieee80211LBMSRequest *frame);
    virtual void handleLbmsReportFrame(Ieee80211LBMSReport *frame);
    virtual void handleDataFrame(Ieee80211DataFrame *frame);
    //@}
};

#endif

