//
// Copyright (c) 2007 INRIA
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2 as
// published by the Free Software Foundation;
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Authors:   Ahmed Ayadi < Ahmed.Ayadi@sophia.inria.fr>
//
#include <itpp/itbase.h>
#include <itpp/itcomm.h>

#include "ShadowingModel.h"


Register_Class(ShadowingModel);

void ShadowingModel::initializeFrom(cModule *radioModule)
{
    EV<< "ShadowingModel selected "<< endl;
    systemLoss = (double) radioModule->par("systemLoss");
    pathLossExponent = (double) radioModule->par("pathLossExponent");
    receiverAntennaGain = (double) radioModule->par("receiverAntennaGain");
    shadowingVariance = (double) radioModule->par("shadowingVariance");
    shadowingNumberofSamples = (double)  radioModule->par("shadowingNumberofSamples");

    shadowingRandomNumberVectorIndex = 0;
    itpp::Normal_RNG * randClass = new itpp::Normal_RNG(0 , shadowingVariance);
    shadowingRandomNumberVector = randClass->operator()(shadowingNumberofSamples);
}

double ShadowingModel::calculateReceivedPower(double pSend, double carrierFrequency, double distance)
{
	double power;	
    const double speedOfLight = 300000000.0;
    double waveLength = speedOfLight / carrierFrequency;
	if ( distance < 1.0)
	{
		power=pSend * waveLength * waveLength / (16 * M_PI * M_PI * pow(distance,2));
	}
	else
	{
		double prd0 = pSend * waveLength * waveLength / (16 * M_PI * M_PI * 1.0 * 1.0 );
		if (shadowingRandomNumberVectorIndex < shadowingNumberofSamples)
		{
			shadowing = shadowingRandomNumberVector[shadowingRandomNumberVectorIndex];
			shadowingRandomNumberVectorIndex++;
		}
		else
		{
	        shadowingRandomNumberVectorIndex=0;
		    shadowing = shadowingRandomNumberVector[shadowingRandomNumberVectorIndex];
		}

		//std::cout << " Shadowing " << shadowing;
		double powerLoss_db = -10.0 * pathLossExponent * log10(distance) + shadowing;
		power = prd0 * pow(10.0, powerLoss_db/10.0);
	}
	return power;
}

