//
// Copyright (c) 2007 INRIA
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2 as
// published by the Free Software Foundation;
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// @author Ahmed Ayadi < Ahmed.Ayadi@sophia.inria.fr>
// @date 2008-01-11
//

#ifndef __RTCP_RATE_CONTROL_H__
#define __RTCP_RATE_CONTROL_H__

#include <omnetpp.h>

#include "defs.h"

#include "types.h"
#include "RTPInnerPacket.h"
#include "RTPParticipantInfo.h"
#include "RTPSenderInfo.h"
#include "RTPReceiverInfo.h"
#include "RTCPEndsystemModule.h"

/**
 * The class RTCPCoordinatorModule is responsible for creating, receiving and
 * processing of rtcp packets. It also keeps track of this and other
 * rtp end systems.
 */

class INET_API RTCPCoordinatorModule : public RTCPEndsystemModule
{

  struct MobileReceiverInfo
  {
      u_int32 ssrc;
      double fractionloss;
      double rtt;
	  double lsr;
  };

  double _lsr;
  double _ntpSender;
  double _rtt;
  double _fractionLost;
  int _stategie ;

  typedef std::vector<MobileReceiverInfo> MobileReceiverInfoList;
  typedef std::vector<MobileReceiverInfo>::iterator MobileReceiverInfoListI;

  MobileReceiverInfoList mobileReceiverInfoList;

  protected:
        /**
         * Initializes variables.
         */
        virtual void initialize();

		virtual ~RTCPCoordinatorModule();
        /**
         * Handles self messages.
         */
        virtual void handleSelfMessage(cMessage *msg);

        /**
         * Called when the socket layer has finished a connect.
         */
        virtual void connectRet();

		void myMobileReceiverInfo();

		void removeMobileReceiver(u_int32 ssrc);

		void addNewMobileReceiver(u_int32 ssrc);

		bool isNewMobileReceiver(u_int32 ssrc);

		double calculateRoundTripTime(ReceptionReport *receptionReport);

		double calculatePacketLoss(ReceptionReport *receptionReport);

		void processMobileReceiverReport(RTCPReceiverReportPacket *rtcpReceiverReportPacket);

		void computeWorstCondition();
		void computeBestCondition();

		void computeMediumCondition();

    private:

        /**
         * The bandwidth for this rtp session.
         */
        int _bandwidth;

		/**
		 * Tha reduction factor
		 */
		int _reductionfactor;

        /**
         * The percentage of bandwidth for rtcp.
         */
        int _rtcpPercentage;

        /**
         * The number of packets this rtcp module has
         * calculated.
         */
        int _packetsCalculated;

        /**
         * The average size of an RTCPCompoundPacket.
         */
        double _averagePacketSize;

		/**
         * The number of rtp packets sent
         */
	    int _rtpPacketsSent;

		/**
         * The average size of RTP packets sent by an RTP module.
         */
		double _rtpAveragePacketSize;

        /**
         * Creates and sends an RTCPCompoundPacket.
         */
        virtual void createPacket();

        /**
         * Extracts information of a sent RTPPacket.
         */
        virtual void processOutgoingRTPPacket(RTPPacket *packet);

        /**
         * Extracts information of a received RTCPCompoundPacket.
         */
        virtual void processIncomingRTCPPacket(RTCPCompoundPacket *packet, IN_Addr address, IN_Port port);
};

#endif


