//
// Copyright (c) 2008 INRIA
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2 as
// published by the Free Software Foundation;
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// @author Ahmed Ayadi < Ahmed.Ayadi@sophia.inria.fr>
// @date 2008-01-11
///

#include <omnetpp.h>

#include "RTPPayloadReceiver.h"
#include "RTPAVProfilePayload100Receiver.h"
#include "RTPPacket.h"
#include "RTPMpegPacket.h"
#include "RTP264Packet_m.h"

Define_Module_Like(RTPAVProfilePayload100Receiver, RTPPayloadReceiver);

RTPAVProfilePayload100Receiver::~RTPAVProfilePayload100Receiver()
{
};


void RTPAVProfilePayload100Receiver::initialize() {
    RTPPayloadReceiver::initialize();
    _payloadType = 100;
    _lowestAllowedTimeStamp = 0;
    _highestSequenceNumber = 0;
    _lowestSequenceNumber = 0;
	compte = 0;
	lastrcvd = 0;
	fpClient = fopen(par("outputFileName"),"wb");

    WATCH(_lowestAllowedTimeStamp);
    WATCH(_highestSequenceNumber);
    WATCH(_lowestSequenceNumber);
};


void RTPAVProfilePayload100Receiver::processPacket(RTPPacket *rtpPacket)
{
	char *data;
	unsigned size;
    // the first packet sets the lowest allowed time stamp
    if (_lowestAllowedTimeStamp == 0)
    {
        _lowestAllowedTimeStamp = rtpPacket->timeStamp();
        _lowestSequenceNumber = rtpPacket->sequenceNumber();
    }

    if ((rtpPacket->timeStamp() < _lowestAllowedTimeStamp) || (rtpPacket->sequenceNumber()<= _highestSequenceNumber)) {
        delete rtpPacket;
    }
    else {
        // is this packet from the next frame ?
        // this can happen when the marked packet has been
        // lost or arrives late
        bool nextTimeStamp = rtpPacket->timeStamp() > _lowestAllowedTimeStamp;

        // is this packet marked, which means that it's
        // the last packet of this frame
        bool marked = rtpPacket->marker();

        // test if end of frame reached

        // check if we received the last (= marked)
        // packet of the frame or
        // we received a packet of the next frame
        _highestSequenceNumber = rtpPacket->sequenceNumber();

        int frameSize = 0;

        // the queue contains all packets for this frame
        // we have received
        RTP264Packet *mpegPacket = (RTP264Packet *)(rtpPacket->decapsulate());

		size = mpegPacket->getDataArraySize();

		data = (char *)malloc(size * sizeof(char));
		for(int i=0;i<size;i++)
		{
			data[i] = mpegPacket->getData(i);
		}
		compte += size;

		fwrite(data,size,1,fpClient);
		free(data);

        delete mpegPacket;
        delete rtpPacket;
		
    };
};


void RTPAVProfilePayload100Receiver::finish()
{
	fclose(fpClient);
}	