//
// Copyright (c) 2007 INRIA
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2 as
// published by the Free Software Foundation;
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// @author Ahmed Ayadi < Ahmed.Ayadi@sophia.inria.fr>
// @date 2008-01-11
//

#include <string.h>
#include <cassert>
#include <omnetpp.h>

#include "RTPPacket.h"
#include "RTPInnerPacket.h"
#include "RTPAVProfilePayload100Sender.h"
#include "RTPMpegPacket.h"
#include "RTP264Packet_m.h"

Define_Module_Like(RTPAVProfilePayload100Sender, RTPPayloadSender);

void RTPAVProfilePayload100Sender::initialize() {

    RTPPayloadSender::initialize();

    // bandwidth in bytes per second for this session
    _bandwidth = par("bandwidth");

	//std::cout << " BandWidth " << _bandwidth <<endl;
    _clockRate = 90000;
    _payloadType = 100;
	
	offset = 0;
	offdeb = 0;
	offfin = 0;
	size = 0;
	comptesize = 0;

	//fpServeur = fopen(par("fileName"),"rb");
};


void RTPAVProfilePayload100Sender::initializeSenderModule(RTPInnerPacket *rinpIn)
{
	fpServeur = fopen(rinpIn->fileName(),"rb");
	
    RTPPayloadSender::initializeSenderModule(rinpIn);

	fseek(fpServeur,0,SEEK_END);
	fin = ftell(fpServeur);
	rewind(fpServeur);
	

	offset = ftell(fpServeur);
	offdeb = offset + 5;
	fread(label,4,1,fpServeur);

	fseek(fpServeur,offdeb,SEEK_SET);

	RTPPacket *rtpPacket = new RTPPacket("RTPPacket");
	RTP264Packet *mpegPacket = new RTP264Packet("MpegPacket");
		
	mpegPacket->setDataArraySize(4);

	for(int j=0;j<4;j++)
	{
		mpegPacket->setData(j,label[j]);//envoi du label.
	}

	mpegPacket->addLength(size);

	rtpPacket->encapsulate(mpegPacket);

	rtpPacket->setPayloadType(_payloadType);
	rtpPacket->setSequenceNumber(_sequenceNumber);
	_sequenceNumber++;

	rtpPacket->setTimeStamp(_timeStampBase + ((simTime() - _timeStampBase) * _clockRate));
	rtpPacket->setSSRC(_ssrc);

	RTPInnerPacket *rinpOut = new RTPInnerPacket("dataOut()");

	rinpOut->dataOut(rtpPacket);

	send(rinpOut, "toProfile");

	_frameNumber++;

	// wait initial delay
    cMessage *reminderMessage = new cMessage("next frame");
    scheduleAt(simTime() + _initialDelay, reminderMessage);

};


bool RTPAVProfilePayload100Sender::sendPacket()
{
	char *data;

	if ((!feof(fpServeur))&&(!TestEndFile(this,fin)))
	{
		fread(mot,4,1,fpServeur);
		offset = ftell(fpServeur);

		while(Strcmp(mot,label))
		{
			if (offset > (fin - 4))
			{
				offset += 4;
				break;
			}

			fread(mot,4,1,fpServeur);
			offset = ftell(fpServeur);
			offset -= 3;
			fseek(fpServeur,offset,SEEK_SET);
		}
		offfin = offset;
			
		size = offfin - offdeb;
		comptesize += size;
		data = (char *)malloc(size);
		fseek(fpServeur,offdeb-1,SEEK_SET);

		fread(data,size,1,fpServeur);
		fseek(fpServeur,offset,SEEK_SET);
		offdeb = offset;
			
		RTPPacket *rtpPacket = new RTPPacket("RTPPacket");
		RTP264Packet *mpegPacket = new RTP264Packet("MpegPacket");
		
		mpegPacket->setDataArraySize(size);

		for(int i=0;i<size;i++)
		{
			mpegPacket->setData(i,data[i]);
		}
		
		fseek(fpServeur,offset,SEEK_SET);
		free(data);

		mpegPacket->addLength(size);

		rtpPacket->encapsulate(mpegPacket);

		rtpPacket->setPayloadType(_payloadType);
		rtpPacket->setSequenceNumber(_sequenceNumber);
		_sequenceNumber++;

		rtpPacket->setTimeStamp(_timeStampBase + ((simTime() - _timeStampBase) * _clockRate));
		rtpPacket->setSSRC(_ssrc);

		RTPInnerPacket *rinpOut = new RTPInnerPacket("dataOut()");

		rinpOut->dataOut(rtpPacket);

		send(rinpOut, "toProfile");

		_frameNumber++;

		_reminderMessage = new cMessage("nextFrame");
		double nextFrame = (double)size/(double)_bandwidth;

		scheduleAt(simTime() + nextFrame, _reminderMessage);
		return true;
    }
    else {
		size = 0;
		RTPPacket *rtpPacket = new RTPPacket("RTPPacket");
		RTP264Packet *mpegPacket = new RTP264Packet("MpegPacket");
		if (offdeb < fin)
		{
			size = fin - offdeb;
			comptesize += size;
			data = (char *)malloc(size);
			fseek(fpServeur,offdeb,SEEK_SET);
			fread(data,size,1,fpServeur);
		
			mpegPacket->setDataArraySize(size);
			std::cout << "Last_Sender_Length_100 : " << size << endl;
			for(int i=0;i<size;i++)	mpegPacket->setData(i,data[i]);

			mpegPacket->addLength(size);
			free(data);
		}
		mpegPacket->addLength(size);
		rtpPacket->encapsulate(mpegPacket);
		rtpPacket->setPayloadType(_payloadType);
		rtpPacket->setSequenceNumber(_sequenceNumber);
		_sequenceNumber++;

		rtpPacket->setTimeStamp(_timeStampBase + ((simTime() - _timeStampBase) * _clockRate));
		rtpPacket->setSSRC(_ssrc);

        RTPInnerPacket *rinpOut = new RTPInnerPacket("dataOut()");
		rinpOut->dataOut(rtpPacket);
        rinpOut->leaveSession();
        send(rinpOut, "toProfile");
        return false;
    }
};


int RTPAVProfilePayload100Sender::TestEndFile(RTPAVProfilePayload100Sender *pt,int fin)
{ 
	int debut = ftell(pt->fpServeur);
	if((fin - debut)<=4) return 1;
	else return 0;
}

void RTPAVProfilePayload100Sender::finish()
{
	fclose(fpServeur);
}	


int RTPAVProfilePayload100Sender::Strcmp(char mot[4],char label[4])
{
	char *ptmot = mot;
	char *ptlabel = label;

	if((ptmot[0]==ptlabel[0])&&
		(ptmot[1]==ptlabel[1])&&
		(ptmot[2]==ptlabel[2])&&
		(ptmot[3]==ptlabel[3]))
	return 0;
	else return 1;
}
