/***************************************************************************
                          RTPEndsystemModule.cc  -  description
                             -------------------
    begin                : Fri Oct 19 2001
    copyright            : (C) 2001 by Matthias Oppitz
    email                : Matthias.Oppitz@gmx.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/** \file RTPEndsystemModule.cc
 * This file contains the implementation of member functions of the class RTPEndsystemModule.
 */

#include <omnetpp.h>

#include "defs.h"
#include "IPAddress.h"
#include "UDPSocket.h"
#include "UDPControlInfo_m.h"
#include "IGMPMessage_m.h"

#include "RTPEndsystemModule.h"
#include "RTPInterfacePacket.h" 
#include "RTPInnerPacket.h"
#include "RTPProfile.h"

#include "RTPSenderControlMessage.h"
#include "RTPSenderStatusMessage.h"

#define std_out

Define_Module_Like(RTPEndsystemModule, RTPModule);


//
// methods inherited from cSimpleModule
//

void RTPEndsystemModule::initialize()
{
    _socketFdIn = -1;
    _socketFdOut = -1;
    _leaveSession = false;

	r_totalSize = r_index = 0;
	lastArrivalTime = 0;

	for (int j=0; j<WIN_SIZE; j++)
	{
		r_paquet[j].arrivalTime =0.0;
		r_paquet[j].size = 0;
	}

	tabIndex = 0;
	intervalSize = totalSize = 0;
	for (int k=0; k<TAB_SIZE; k++)
	{
		_intervalSizes[k] = 0;
	}
	_receivedThroughputwithFilter.setName("Received Throughput with Filter");
	_sentThroughput.setName("Sent Throughput");

	cMessage *recordMessage = new cMessage("Record Interval");
    scheduleAt(simTime() + WIN_DELTA, recordMessage);
};

void RTPEndsystemModule::handleMessage(cMessage *msg)
{
    if (msg->arrivalGateId() == findGate("fromApp")) {
        handleMessageFromApp(msg);
    }

    else if (msg->arrivalGateId() == findGate("fromProfile")) {
        handleMessageFromProfile(msg);
    }

    else if (msg->arrivalGateId() == findGate("fromRTCP")) {
        handleMessageFromRTCP(msg);
    }

    else if (msg->arrivalGateId() == findGate("fromUDPLayer")) {
        handleMessagefromUDP(msg);
    }
	else if (msg->isSelfMessage()) {
        recordReceivedRate(msg);
    }
    else {
        ev << " Message from unknown Gate !" << endl;
    }
};


//
// handle messages from different gates
//

void RTPEndsystemModule::handleMessageFromApp(cMessage *msg)
{
    RTPInterfacePacket *rifp = (RTPInterfacePacket *)msg;
    if (rifp->type() == RTPInterfacePacket::RTP_IFP_ENTER_SESSION) {
        enterSession(rifp);
    }
    else if (rifp->type() == RTPInterfacePacket::RTP_IFP_CREATE_SENDER_MODULE) {
        createSenderModule(rifp);
    }
    else if (rifp->type() == RTPInterfacePacket::RTP_IFP_DELETE_SENDER_MODULE) {
        deleteSenderModule(rifp);
    }
    else if (rifp->type() == RTPInterfacePacket::RTP_IFP_SENDER_CONTROL) {
        senderModuleControl(rifp);
    }
    else if (rifp->type() == RTPInterfacePacket::RTP_IFP_LEAVE_SESSION) {
        leaveSession(rifp);
    }
    else {
        ev << " unknown RTPInterfacePacket type from application !" << endl;
    }
};


void RTPEndsystemModule::handleMessageFromProfile(cMessage *msg)
{
    RTPInnerPacket *rinp = (RTPInnerPacket *)msg;
    if (rinp->type() == RTPInnerPacket::RTP_INP_PROFILE_INITIALIZED) {
        profileInitialized(rinp);
    }
    else if (rinp->type() == RTPInnerPacket::RTP_INP_SENDER_MODULE_CREATED) {
        senderModuleCreated(rinp);
    }
    else if (rinp->type() == RTPInnerPacket::RTP_INP_SENDER_MODULE_DELETED) {
        senderModuleDeleted(rinp);
    }
    else if (rinp->type() == RTPInnerPacket::RTP_INP_SENDER_MODULE_INITIALIZED) {
        senderModuleInitialized(rinp);
    }
    else if (rinp->type() == RTPInnerPacket::RTP_INP_SENDER_MODULE_STATUS) {
        senderModuleStatus(rinp);
    }
    else if (rinp->type() == RTPInnerPacket::RTP_INP_RECEIVER_MODULE_STATS) {
        receiverModuleStats(rinp);
    }
    else if (rinp->type() == RTPInnerPacket::RTP_INP_LEAVE_SESSION) {
        leaveSession(rinp);
    }
    else if (rinp->type() == RTPInnerPacket::RTP_INP_DATA_OUT) {
        dataOut(rinp);
    }
    else {
        delete msg;
    }
    ev << "handleMessageFromProfile(cMessage *msg) Exit"<<endl;
}


void RTPEndsystemModule::handleMessageFromRTCP(cMessage *msg)
{
    RTPInnerPacket *rinp = (RTPInnerPacket *)msg;
    if (rinp->type() == RTPInnerPacket::RTP_INP_RTCP_INITIALIZED) {
        rtcpInitialized(rinp);
    }
    else if (rinp->type() == RTPInnerPacket::RTP_INP_CHANGE_BANDWIDTH) {
        adaptBitRate(rinp);
    }
	    else if (rinp->type() == RTPInnerPacket::RTP_INP_SESSION_LEFT) {
        sessionLeft(rinp);
    }
    else {
        ev << "Unknown RTPInnerPacket type " << rinp->type() << " from rtcp !" << endl;
    }
};

void RTPEndsystemModule::handleMessagefromUDP(cMessage *msg)
{
    readRet(msg);
};


//
// methods for different messages
//

void RTPEndsystemModule::enterSession(RTPInterfacePacket *rifp)
{
    _profileName = rifp->profileName();
    _commonName = rifp->commonName();
    _bandwidth = rifp->bandwidth();
    _destinationAddress = rifp->destinationAddress();

    _port = rifp->port();
    if (_port % 2 != 0) {
        _port = _port - 1;
    }

    _mtu = resolveMTU();

    createProfile();
    initializeProfile();

    if (_destinationAddress.isMulticast())
		sendJoin();

    delete rifp;
};


void RTPEndsystemModule::leaveSession(cMessage *rifp)
{
    cModule *profileModule = gate("toProfile")->toGate()->ownerModule();
    profileModule->deleteModule();
    _leaveSession = true;
    RTPInnerPacket *rinp = new RTPInnerPacket("leaveSession()");
    rinp->leaveSession();
    send(rinp,"toRTCP");

   if (_destinationAddress.isMulticast())
	   sendLeave();

   delete rifp;
};


void RTPEndsystemModule::createSenderModule(RTPInterfacePacket *rifp)
{
    RTPInnerPacket *rinp = new RTPInnerPacket("createSenderModule()");
    ev << rifp->ssrc()<<endl;
    rinp->createSenderModule(rifp->ssrc(), rifp->payloadType(), rifp->fileName());
    send(rinp, "toProfile");

    delete rifp;
};


void RTPEndsystemModule::deleteSenderModule(RTPInterfacePacket *rifp)
{
    RTPInnerPacket *rinp = new RTPInnerPacket("deleteSenderModule()");
    rinp->deleteSenderModule(rifp->ssrc());
    send(rinp, "toProfile");

    delete rifp;
};


void RTPEndsystemModule::senderModuleControl(RTPInterfacePacket *rifp)
{
    RTPInnerPacket *rinp = new RTPInnerPacket("senderModuleControl()");
    rinp->senderModuleControl(rinp->ssrc(), (RTPSenderControlMessage *)(rifp->decapsulate()));
    send(rinp, "toProfile");

    delete rifp;
};


void RTPEndsystemModule::profileInitialized(RTPInnerPacket *rinp)
{
    _rtcpPercentage = rinp->rtcpPercentage();
    if (_port == IPSuite_PORT_UNDEF) {
        _port = rinp->port();
        if (_port % 2 != 0) {
            _port = _port - 1;
        }
    }

    delete rinp;

    createSocket();
};


void RTPEndsystemModule::senderModuleCreated(RTPInnerPacket *rinp)
{
    RTPInterfacePacket *rifp = new RTPInterfacePacket("senderModuleCreated()");
    rifp->senderModuleCreated(rinp->ssrc());
    send(rifp, "toApp");

    delete rinp;
};


void RTPEndsystemModule::senderModuleDeleted(RTPInnerPacket *rinp)
{
    RTPInterfacePacket *rifp = new RTPInterfacePacket("senderModuleDeleted()");
    rifp->senderModuleDeleted(rinp->ssrc());
    send(rifp, "toApp");

    // perhaps we should send a message to rtcp module
    delete rinp;
};


void RTPEndsystemModule::senderModuleInitialized(RTPInnerPacket *rinp)
{
    send(rinp, "toRTCP");
};


void RTPEndsystemModule::senderModuleStatus(RTPInnerPacket *rinp)
{
    RTPInterfacePacket *rifp = new RTPInterfacePacket("senderModuleStatus()");
    rifp->senderModuleStatus(rinp->ssrc(), (RTPSenderStatusMessage *)(rinp->decapsulate()));
    send(rifp, "toApp");

    delete rinp;
};


void RTPEndsystemModule::dataOut(RTPInnerPacket *rinp)
{
    RTPPacket *msg = (RTPPacket *)(rinp->decapsulate());
	
	recordSentRate(msg->length()-16, simTime());

    // send message to UDP, with the appropriate control info attached
    msg->setKind(UDP_C_DATA);

    UDPControlInfo *ctrl = new UDPControlInfo();
    ctrl->setDestAddr(_destinationAddress);
    ctrl->setDestPort(_port);
    msg->setControlInfo(ctrl);

    send(msg, "toUDPLayer");

    // RTCP module must be informed about sent rtp data packet

    RTPInnerPacket *rinpOut = new RTPInnerPacket(*rinp);
    rinpOut->encapsulate(new RTPPacket(*msg));
    send(rinpOut, "toRTCP");

    delete rinp;
};


void RTPEndsystemModule::rtcpInitialized(RTPInnerPacket *rinp)
{
    RTPInterfacePacket *rifp = new RTPInterfacePacket("sessionEntered()");
    rifp->sessionEntered(rinp->ssrc());
    send(rifp, "toApp");

    delete rinp;
};


void RTPEndsystemModule::sessionLeft(RTPInnerPacket *rinp)
{
    RTPInterfacePacket *rifp = new RTPInterfacePacket("sessionLeft()");
    rifp->sessionLeft();
    send(rifp, "toApp");

    delete rinp;
};


void RTPEndsystemModule::connectRet()
{
    initializeRTCP();
};


void RTPEndsystemModule::readRet(cMessage *sifp)
{
	//std::cout<<"Read ret start" << simTime() << endl;
    if ( ! _leaveSession) 
    {
         RTPPacket *msg = (RTPPacket *)(sifp);

		 intervalSize += msg->length()-16;

         msg->writeContents();
         RTPInnerPacket *rinp1 = new RTPInnerPacket("dataIn1()");
         rinp1->dataIn(new RTPPacket(*msg),IN_Addr(_destinationAddress),IN_Port(_port));

         RTPInnerPacket *rinp2 = new RTPInnerPacket(*rinp1);
         send(rinp2, "toRTCP");
		 if (this->gate("toProfile")->isConnected())
			send(rinp1, "toProfile");
    }

    delete sifp;
	//std::cout<<"Read ret end" << endl;
};


int RTPEndsystemModule::resolveMTU()
{
    // this is not what it should be
    // do something like mtu path discovery
    // for the simulation we can use this example value
    // it's 1500 bytes (ethernet) minus ip
    // and udp headers
    return 1500 - 20 - 8;
};


void RTPEndsystemModule::createProfile()
{
    cModuleType *moduleType = findModuleType(_profileName);
    if (moduleType == NULL) {
        error("Profile type not found !");
    };
    RTPProfile *profile = (RTPProfile *)(moduleType->create("Profile", this));

    profile->setGateSize("toPayloadReceiver", 30);
    profile->setGateSize("fromPayloadReceiver", 30);

    this->gate("toProfile")->connectTo(profile->gate("fromRTP"));
    profile->gate("toRTP")->connectTo(this->gate("fromProfile"));

    profile->initialize();
    profile->scheduleStart(simTime());
};


void RTPEndsystemModule::createSocket()
{
    if (_socketFdIn == -1) {
    _socketFdIn = UDPSocket::generateSocketId();

    UDPControlInfo *ctrl = new UDPControlInfo();

    IPAddress ipaddr(_destinationAddress);

    if (ipaddr.isMulticast()) {
        ctrl->setSrcAddr(IN_Addr(_destinationAddress));
        ctrl->setSrcPort(IN_Port(_port));
    }
    else {
         ctrl->setSrcPort(_port);
         ctrl->setSockId(_socketFdOut);
    }
    ctrl->setSockId((int)_socketFdIn);
    cMessage *msg = new cMessage("UDP_C_BIND", UDP_C_BIND);
    msg->setControlInfo(ctrl);
    send(msg,"toUDPLayer");

    connectRet();
    }
};

void RTPEndsystemModule::initializeProfile()
{
    RTPInnerPacket *rinp = new RTPInnerPacket("initializeProfile()");
    rinp->initializeProfile(_mtu);
    send(rinp, "toProfile");
};


void RTPEndsystemModule::initializeRTCP()
{
    RTPInnerPacket *rinp = new RTPInnerPacket("initializeRTCP()");
    IN_Port rtcpPort = _port + 1;
    rinp->initializeRTCP(opp_strdup(_commonName), _mtu, _bandwidth, _rtcpPercentage, _destinationAddress, rtcpPort);
    send(rinp, "toRTCP");
};

void RTPEndsystemModule::sendJoin()
{
	if (_destinationAddress.isMulticast())
	{
		IGMPMessage *msg = new IGMPMessage("Join");
		msg->setType(MEMBERSHIP_REPORT);
		msg->setGroupAddress(IN_Addr(_destinationAddress));

		// send message to UDP, with the appropriate control info attached
		msg->setKind(UDP_C_DATA);

		UDPControlInfo *ctrl = new UDPControlInfo();
		ctrl->setDestAddr("224.0.0.1");
		ctrl->setDestPort(_port);
		msg->setControlInfo(ctrl);

		send(msg, "toUDPLayer");
	}
};


void RTPEndsystemModule::sendLeave()
{
	if (_leaveSession && _destinationAddress.isMulticast())
    {
        ev<< "Leave session"<<endl;
        IGMPMessage *msg = new IGMPMessage("Leave");
        msg->setType(LEAVE_GROUP);
        msg->setGroupAddress(IN_Addr(_destinationAddress));

        // send message to UDP, with the appropriate control info attached
        msg->setKind(UDP_C_DATA);

        UDPControlInfo *ctrl = new UDPControlInfo();
        ctrl->setDestAddr("224.0.0.1");
        ctrl->setDestPort(_port);
        msg->setControlInfo(ctrl);

        send(msg, "toUDPLayer");
	}
};


void RTPEndsystemModule::receiverModuleStats(RTPInnerPacket *rinp)
{
    // send message to UDP, with the appropriate control info attached
    rinp->setKind(UDP_C_DATA);

    UDPControlInfo *ctrl = new UDPControlInfo();
    ctrl->setDestAddr("224.0.0.1");
    ctrl->setDestPort(_port);
    rinp->setControlInfo(ctrl);

    send(rinp, "toUDPLayer");
//     delete rinp;
};

void RTPEndsystemModule::adaptBitRate(RTPInnerPacket *rinp)
{
	std::cout << " RTPEndsystemModule::adaptBitRate "<< endl;
	if ( ! _leaveSession) 
    {
		if (this->gate("toProfile")->isConnected())
		send(rinp, "toProfile");
	}
	else rinp;
};

void RTPEndsystemModule::recordSentRate(int size, double arrivalTime)
{
	 if (lastArrivalTime > 0)
	 {
		double sentThroughput = ((double)size/(arrivalTime-lastArrivalTime))/1000;
		//_sentThroughput.record(sentThroughput);
	 }
	lastArrivalTime = arrivalTime;
};


void RTPEndsystemModule::recordReceivedRate(cMessage *msg)
{
	int index = tabIndex % TAB_SIZE;
	double rvcdThroughput = 0.0;

	if (tabIndex >= TAB_SIZE)
	{
		double rvcdThroughput = 0.0;
		totalSize = totalSize + intervalSize - _intervalSizes[index];
		rvcdThroughput = totalSize / WIN_TIME_SIZE;
		_receivedThroughputwithFilter.record(rvcdThroughput/1000);
	}

	_intervalSizes[index] = intervalSize;
	intervalSize = 0;
	tabIndex++;
	 
    scheduleAt(simTime() + WIN_DELTA, msg);
};
