//
// Copyright (c) 2007 INRIA
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2 as
// published by the Free Software Foundation;
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// @author Ahmed Ayadi < Ahmed.Ayadi@sophia.inria.fr>
// @date 2008-01-11
//
#include "RTPMpegWithPayloadPacket.h"

Register_Class(RTPMpegWithPayloadPacket);

RTPMpegWithPayloadPacket::RTPMpegWithPayloadPacket(const char *name) : cPacket(name)
{
    data_arraysize = 0;
    this->data_var = 0;
};


void RTPMpegWithPayloadPacket::setDataArraySize(unsigned int size)
{
    char *data_var2 = (size==0) ? NULL : new char[size];
    unsigned int sz = data_arraysize < size ? data_arraysize : size;
    unsigned int i;
    for (i=0; i<sz; i++)
        data_var2[i] = this->data_var[i];
    for (i=sz; i<size; i++)
        data_var2[i] = 0;
    data_arraysize = size;
    delete [] this->data_var;
    this->data_var = data_var2;
};


unsigned int RTPMpegWithPayloadPacket::getDataArraySize() const
{
    return data_arraysize;
};

char RTPMpegWithPayloadPacket::getData(unsigned int k) const
{
    if (k>=data_arraysize) throw new cException("Array of size %d indexed by %d", data_arraysize, k);
    return data_var[k];
};

void RTPMpegWithPayloadPacket::setData(unsigned int k, char data_var)
{
    if (k>=data_arraysize) throw new cException("Array of size %d indexed by %d", data_arraysize, k);
    this->data_var[k]=data_var;
};

RTPMpegWithPayloadPacket::~RTPMpegWithPayloadPacket()
{
    delete [] data_var;
};

int RTPMpegWithPayloadPacket::headerLength() {
    return 4;
};


int RTPMpegWithPayloadPacket::payloadLength() {
    return length() - headerLength();
};

